"""Module contains classes for storing sessions"""

import time
from typing import Callable


class _Missing:
    """Represents a missing value in the dictionary."""

    def __repr__(self) -> str:
        return "no value"

    def __reduce__(self) -> str:
        return "_missing"


_missing = _Missing()


class UpdateDictMixin[T, P]:
    """Mixin class that triggers a callback on dictionary updates."""

    @staticmethod
    def calls_update(name: str):
        """Decorator to trigger `on_update` callback after a dictionary operation.

        Args:
            name (str): The name of the dictionary method to wrap.

        Returns:
            Callable: The wrapped method.
        """

        def oncall(self, *args, **kwargs):
            rv = getattr(super(), name)(*args, **kwargs)
            if self.on_update is not None:
                self.on_update(self)
            return rv

        oncall.__name__ = name
        return oncall

    def setdefault(self, key: T, default: P | None = None) -> P:
        """Sets a default value for a key if it does not exist.

        Args:
            key (T): The key to set.
            default (P | None): The default value to set.

        Returns:
            P: The value of the key.
        """
        modified = key not in self
        rv = super().setdefault(key, default)
        if modified and self.on_update is not None:
            self.on_update(self)
        return rv

    def pop(self, key: T, default: P | None = _missing) -> P:
        """Removes a key from the dictionary.

        Args:
            key (T): The key to remove.
            default (P | None): The default value to return if the key is missing.

        Returns:
            P: The value of the removed key or the default value.
        """
        modified = key in self
        if default is _missing:
            rv = super().pop(key)
        else:
            rv = super().pop(key, default)
        if modified and self.on_update is not None:
            self.on_update(self)
        return rv

    __setitem__ = calls_update("__setitem__")
    __delitem__ = calls_update("__delitem__")
    clear = calls_update("clear")
    popitem = calls_update("popitem")
    update = calls_update("update")
    del calls_update


class CallbackDict[T, P](UpdateDictMixin[T, P], dict):
    """A dictionary that triggers a callback on updates."""

    def __init__(self, initial: dict[T, P] | None = None, on_update: Callable | None = None):
        """Initializes the CallbackDict.

        Args:
            initial (dict[T, P] | None): Initial dictionary values.
            on_update (Callable | None): Callback function to trigger on updates.
        """
        dict.__init__(self, initial or {})
        self.on_update = on_update

    def __repr__(self) -> str:
        """Returns a string representation of the dictionary.

        Returns:
            str: The string representation.
        """
        return f"<{self.__class__.__name__} {dict.__repr__(self)}>"


class ExpiringDict[T, P](dict):
    """A dictionary with expiring keys."""

    def __init__(self, prefix: str = ""):
        """Initializes the ExpiringDict.

        Args:
            prefix (str): A prefix to prepend to keys.
        """
        self.prefix: str = prefix
        super().__init__()
        self.expiry_times: dict[T, float] = {}

    def set(self, key: T, val: P, expiry: int) -> None:
        """Sets a key-value pair with an expiry time.

        Args:
            key (T): The key to set.
            val (P): The value to set.
            expiry (int): The expiry time in seconds.
        """
        self[key] = val
        self.expiry_times[key] = time.time() + expiry

    def get_by_sid(self, key: T) -> P | None:
        """Gets a value by its session ID (key with prefix).

        Args:
            key (T): The session ID.

        Returns:
            P | None: The value associated with the key, or None if expired or missing.
        """
        key = self.prefix + str(key)
        return self.get(key)

    def get(self, key: T) -> P | None:
        """Gets a value by its key.

        Args:
            key (T): The key to retrieve.

        Returns:
            P | None: The value associated with the key, or None if expired or missing.
        """
        data = dict(self).get(key)

        if not data:
            return None

        if time.time() > self.expiry_times[key]:
            del self[key]
            del self.expiry_times[key]
            return None

        return data

    def delete(self, key: T) -> None:
        """Deletes a key from the dictionary.

        Args:
            key (T): The key to delete.
        """
        del self[key]
        del self.expiry_times[key]
