"""Module with Sessions"""

import abc
import datetime
import time
import uuid

import ujson
from sanic import Sanic
from sanic.request import Request
from sanic.response import HTTPResponse

from .expire_dict import CallbackDict, ExpiringDict


def get_request_container(request: Request) -> dict[str, object]:
    """Gets the container for storing session data in the request.

    Args:
        request (Request): The Sanic request object.

    Returns:
        dict[str, object]: The container for session data.
    """
    return request.ctx.__dict__ if hasattr(request, "ctx") else request


class SessionDict[T, P](CallbackDict[str, P]):
    """A dictionary-like object for storing session data."""

    def __init__(self, initial: dict[str, P] | None = None, sid: str | None = None):
        """Initializes the SessionDict.

        Args:
            initial (dict[str, V] | None): Initial session data.
            sid (str | None): The session ID.
        """

        def on_update(self):
            self.modified = True

        super().__init__(initial, on_update)

        self.sid: str | None = sid
        self.modified: bool = False


class BaseSessionInterface[T, P](metaclass=abc.ABCMeta):
    """Abstract base class for session interfaces."""

    def __init__(
        self,
        expiry: int,
        prefix: str,
        cookie_name: str,
        domain: str | None,
        httponly: bool,
        sessioncookie: bool,
        samesite: str | None,
        session_name: str,
        secure: bool,
    ):
        """Initializes the BaseSessionInterface.

        Args:
            expiry (int): The session expiry time in seconds.
            prefix (str): The prefix for session keys.
            cookie_name (str): The name of the session cookie.
            domain (str | None): The domain for the session cookie.
            httponly (bool): Whether the cookie is HTTP-only.
            sessioncookie (bool): Whether the cookie is a session cookie.
            samesite (str | None): The SameSite attribute for the cookie.
            session_name (str): The name of the session.
            secure (bool): Whether the cookie is secure.
        """
        self.expiry: int = expiry
        self.prefix: str = prefix
        self.cookie_name: str = cookie_name
        self.domain: str | None = domain
        self.httponly: bool = httponly
        self.sessioncookie: bool = sessioncookie
        self.samesite: str | None = samesite
        self.session_name: str = session_name
        self.secure: bool = secure

    def _delete_cookie(self, request: Request, response: HTTPResponse) -> None:
        """Deletes the session cookie.

        Args:
            request (Request): The Sanic request object.
            response (HTTPResponse): The Sanic response object.
        """
        req = get_request_container(request)
        response.cookies.add_cookie(
            self.cookie_name,
            req[self.session_name].sid,
            domain=self.domain,
            secure=self.secure,
            max_age=0,
            expires=datetime.datetime.now(datetime.timezone.utc),
            samesite=self.samesite,
        )

    @staticmethod
    def _calculate_expires(expiry: int) -> datetime.datetime:
        """Calculates the expiry time for the session.

        Args:
            expiry (int): The expiry time in seconds.

        Returns:
            datetime.datetime: The calculated expiry time.
        """
        expires = time.time() + expiry
        return datetime.datetime.fromtimestamp(expires, tz=datetime.timezone.utc)

    def _set_cookie_props(self, request: Request, response: HTTPResponse) -> None:
        """Sets the properties of the session cookie.

        Args:
            request (Request): The Sanic request object.
            response (HTTPResponse): The Sanic response object.
        """
        req = get_request_container(request)
        cookie_value = req[self.session_name].sid
        cookie_params = {
            "httponly": self.httponly,
            "secure": self.secure,
            "domain": self.domain,
            "samesite": self.samesite,
        }
        if not self.sessioncookie:
            cookie_params["expires"] = self._calculate_expires(self.expiry)
            cookie_params["max_age"] = self.expiry
        response.cookies.add_cookie(self.cookie_name, cookie_value, **cookie_params)

    @abc.abstractmethod
    async def _get_value(self, prefix: str, sid: str) -> P | None:
        """Gets a value from the session store.

        Args:
            prefix (str): The prefix for the session key.
            sid (str): The session ID.

        Returns:
            P | None: The session value, or None if not found.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def _delete_key(self, key: str) -> None:
        """Deletes a key from the session store.

        Args:
            key (str): The key to delete.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def _set_value(self, key: str, data: P) -> None:
        """Sets a value in the session store.

        Args:
            key (str): The key to set.
            data (P): The session data to store.
        """
        raise NotImplementedError

    async def open(self, request: Request) -> SessionDict[P]:
        """Opens a session for the given request.

        Args:
            request (Request): The Sanic request object.

        Returns:
            SessionDict[P]: The session data.
        """
        sid = request.cookies.get(self.cookie_name)

        if not sid:
            sid = uuid.uuid4().hex
            session_dict = SessionDict(sid=sid)
        else:
            val = await self._get_value(self.prefix, sid)

            if val is not None:
                data = ujson.loads(val)
                session_dict = SessionDict(data, sid=sid)
            else:
                session_dict = SessionDict(sid=sid)

        req = get_request_container(request)
        req[self.session_name] = session_dict

        return session_dict

    async def save(self, request: Request, response: HTTPResponse) -> None:
        """Saves the session data to the session store.

        Args:
            request (Request): The Sanic request object.
            response (HTTPResponse): The Sanic response object.
        """
        req = get_request_container(request)
        if self.session_name not in req:
            return

        key = self.prefix + req[self.session_name].sid
        if not req[self.session_name]:
            await self._delete_key(key)

            if req[self.session_name].modified:
                self._delete_cookie(request, response)
            return

        val = ujson.dumps(dict(req[self.session_name]))
        await self._set_value(key, val)
        self._set_cookie_props(request, response)


class InMemorySessionStore(BaseSessionInterface[str, str]):
    """An in-memory implementation of the session store."""

    def __init__(
        self,
        domain: str | None = None,
        expiry: int = 2592000,
        httponly: bool = True,
        cookie_name: str = "session",
        prefix: str = "session:",
        sessioncookie: bool = False,
        samesite: str | None = None,
        session_name: str = "session",
        secure: bool = False,
    ):
        """Initializes the InMemorySessionStore.

        Args:
            domain (str | None): The domain for the session cookie.
            expiry (int): The session expiry time in seconds.
            httponly (bool): Whether the cookie is HTTP-only.
            cookie_name (str): The name of the session cookie.
            prefix (str): The prefix for session keys.
            sessioncookie (bool): Whether the cookie is a session cookie.
            samesite (str | None): The SameSite attribute for the cookie.
            session_name (str): The name of the session.
            secure (bool): Whether the cookie is secure.
        """
        super().__init__(
            expiry=expiry,
            prefix=prefix,
            cookie_name=cookie_name,
            domain=domain,
            httponly=httponly,
            sessioncookie=sessioncookie,
            samesite=samesite,
            session_name=session_name,
            secure=secure,
        )
        self.session_store: ExpiringDict[str, str] = ExpiringDict()

    async def _get_value(self, prefix: str, sid: str) -> str | None:
        """Gets a value from the in-memory session store.

        Args:
            prefix (str): The prefix for the session key.
            sid (str): The session ID.

        Returns:
            str | None: The session value, or None if not found.
        """
        return self.session_store.get(self.prefix + sid)

    async def _delete_key(self, key: str) -> None:
        """Deletes a key from the in-memory session store.

        Args:
            key (str): The key to delete.
        """
        if key in self.session_store:
            self.session_store.delete(key)

    async def _set_value(self, key: str, data: str) -> None:
        """Sets a value in the in-memory session store.

        Args:
            key (str): The key to set.
            data (str): The session data to store.
        """
        self.session_store.set(key, data, self.expiry)


class Session:
    """Manages session middleware for a Sanic application."""

    __slots__ = ("interface",)

    def __init__(self, app: Sanic):
        """Initializes the Session middleware.

        Args:
            app (Sanic): The Sanic application instance.
        """
        self.interface: InMemorySessionStore = InMemorySessionStore()

        if not hasattr(app.ctx, "extensions"):
            app.ctx.extensions = {}

        app.ctx.extensions[self.interface.session_name] = self
        app.sanic.register_middleware(self.add_session_to_request, "request")
        app.sanic.register_middleware(self.save_session, "response")

    async def add_session_to_request(self, request: Request) -> None:
        """Adds session data to the request.

        Args:
            request (Request): The Sanic request object.
        """
        await self.interface.open(request)

    async def save_session(self, request: Request, response: HTTPResponse) -> None:
        """Saves session data after the response is processed.

        Args:
            request (Request): The Sanic request object.
            response (HTTPResponse): The Sanic response object.
        """
        await self.interface.save(request, response)
